local Prop = require("gfx.prop")
local Children = Prop:new{ defaultState = {} }

function Children:with(elementType)
  local propSelf = self
  function elementType:_iterateChildren(opts)
    local func, tbl, start = ipairs(propSelf:getState(self))
    return function(t, i)
        if i == opts.to then
            return nil, nil
        end
        return func(t, i)
    end, tbl, (opts and opts.from) or start
  end

  function elementType:childCount()
    return #propSelf:getState(self)
  end

  function elementType:_childAt(index)
    return propSelf:getState()[elementType:_validChildIndex(index)]
  end

  function elementType:_validChildIndex(from)
    return math.max(1, math.min((from or self:childCount()) + 1, self:childCount() + 1))
  end

  function elementType:_triggerChildChanged(index)
    if type(self.onChildrenChanged) == "function" then
        self:onChildrenChanged(index)
    end
    self:_reload()
  end

  function elementType:_addChild(child, index)
    index = self:_validChildIndex(index)
    table.insert(propSelf:getState(self), index, child)
    self:_triggerChildChanged(index)
  end

  function elementType:_removeChild(locator)
    local index = nil
    local locatorType = type(locator)
    if locatorType == "table" then
      for i,v in self:_iterateChildren() do
        if v == locator then
          index = i
          break
        end
      end
    elseif locatorType == "string" then
      for i,v in self:_iterateChildren() do
        if v:getId() == locator then
          index = i
          break
        end
      end
    else--if locatorType == "number" then
      index = self:_validChildIndex(locator)
    end

    -- Child not found
    if index == nil then
        return false, nil
    end

    local result = table.remove(propSelf:getState(self), index)
    self:_triggerChildChanged(index)
    return true, result
  end

  function elementType:findChildById(id)
    for _,child in self:_iterateChildren() do
      local result = child:findById(id)
      if result then
          return result
      end
    end
    return nil
  end

  return elementType
end

return Children